// AllocateComponentsDlg.cpp : implementation file
//

#include "stdafx.h"
#include "Avionics.h"
#include "AllocateComponentsDlg.h"
#include "AIFRead.h"
#include "MainFrm.h"
#include "AvionicsDoc.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CAllocateComponentsDlg dialog


CAllocateComponentsDlg::CAllocateComponentsDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CAllocateComponentsDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CAllocateComponentsDlg)
	m_strAllocResults = _T("");
	m_nCPUs = 1;
	m_nNumComponents = 1000;
	m_nUtil = 100;
	//}}AFX_DATA_INIT
}


void CAllocateComponentsDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CAllocateComponentsDlg)
	DDX_Control(pDX, IDOK, m_btnUpdateAIF);
	DDX_Control(pDX, IDC_COMBO_HEURISTIC, m_cboHeuristic);
	DDX_Text(pDX, IDC_EDIT_RESULTS, m_strAllocResults);
	DDX_Text(pDX, IDC_EDIT_NUMCPUS, m_nCPUs);
	DDX_Text(pDX, IDC_EDIT_MAXNUMCOMPS, m_nNumComponents);
	DDX_Text(pDX, IDC_EDIT_MAXUTIL, m_nUtil);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CAllocateComponentsDlg, CDialog)
	//{{AFX_MSG_MAP(CAllocateComponentsDlg)
	ON_BN_CLICKED(IDC_BUTTON_ALLOCATE, OnButtonAllocate)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CAllocateComponentsDlg message handlers
bool bNoFeedback = false;

void CAllocateComponentsDlg::OnButtonAllocate() 
{
	UpdateData(TRUE);
	double maxUtil = ((double) m_nUtil)/100;
	double totalUtil = 0.0;
	POSITION pos = gTheSystem.compNodeList.GetHeadPosition();
	while(pos) {
		CComponentNode *cNode = gTheSystem.compNodeList.GetNext(pos);
		totalUtil += cNode->util;
	}

	int numCPUs = gTheSystem.processorNodeList.GetCount();
	if(m_nCPUs < 1) {
		AfxMessageBox("Must have total number of CPUs >= 1!");
		return;
	}

	if(m_nCPUs != numCPUs) {
		CString msg;
		msg.Format("The system currently has %d CPUs, while you are requesting %d CPUs for allocation.\
Feedback to AIF file is not possible if they are different.", numCPUs, m_nCPUs);
		bNoFeedback = true;
		AfxMessageBox(msg);
	} else {
		bNoFeedback = false;
	}

	numCPUs = m_nCPUs;
	if(totalUtil > numCPUs) {
		CString msg;
		msg.Format("The total system utilization %.2lf is larger than MaxUtil(%d)*numProcessors (%d). Cannot allocate.",
			totalUtil, m_nUtil, numCPUs);
		AfxMessageBox(msg);
		return;
	}

	//Construct a new list of processors, with all the current processors plus the extra ones
	gTheSystem.newProcessorNodeList.RemoveAll();
	pos = gTheSystem.processorNodeList.GetHeadPosition();
	int count = 0;
	while(pos) {
		CProcessorNode *pNode = gTheSystem.processorNodeList.GetNext(pos);
		gTheSystem.newProcessorNodeList.AddTail(pNode);
		count++;
	}

	while(count < numCPUs) {
		CString name;
		name.Format("NewCPU%d", count++);
		gTheSystem.newProcessorNodeList.AddTail(new CProcessorNode(name, CString("")));
	}

	//First remove all component allocations and start clean.
	//components_allocation is the scratchpad where I put allocation results
	POSITION pos2 = gTheSystem.newProcessorNodeList.GetHeadPosition();
	while(pos2) {
		CProcessorNode *pNode = gTheSystem.newProcessorNodeList.GetNext(pos2);
		pNode->components_allocation.RemoveAll();
	}

	pos = gTheSystem.newProcessorNodeList.GetHeadPosition();
	totalUtil = 0.0;
	Node *n = gTheSystem.theCEDG.BFSSortedNodes.get_head();
	if(m_cboHeuristic.GetCurSel() == 1) {
		//First fit
		while(pos) {
			CProcessorNode *pNode = gTheSystem.newProcessorNodeList.GetNext(pos);
			while(n) {
				CComponentNode *cNode = gTheSystem.theCEDG.get_comp_node(n->x);
				if(cNode->locked && cNode->processor) {
					cNode->processor->AddComponentAllocation(cNode);
				} else if(totalUtil + cNode->util > maxUtil) {
					break;
				} else {
					pNode->AddComponentAllocation(cNode);
					totalUtil += cNode->util;
				}
				n = gTheSystem.theCEDG.BFSSortedNodes.get_next(n);
				
			}
		}
	} else {
		//Best fit
		int numCompsPerCPU = gTheSystem.compNodeList.GetCount() / m_nCPUs;
		while(pos) {
			CProcessorNode *pNode = gTheSystem.newProcessorNodeList.GetNext(pos);
			double totalUtil = 0.0;
			int count = 0;
			while(n) {
				CComponentNode *cNode = gTheSystem.theCEDG.get_comp_node(n->x);
				if(cNode->locked && cNode->processor) {
					cNode->processor->AddComponentAllocation(cNode);
				} else if(totalUtil + cNode->util > maxUtil || count++ > numCompsPerCPU) {
					break;
				} else {
					pNode->AddComponentAllocation(cNode);
					totalUtil += cNode->util;
				}
				n = gTheSystem.theCEDG.BFSSortedNodes.get_next(n);
			}
		}
	}
	
	CProcessorNode *pNode = gTheSystem.newProcessorNodeList.GetTail();
	while(n) {
		CComponentNode *cNode = gTheSystem.theCEDG.get_comp_node(n->x);
		pNode->AddComponentAllocation(cNode);
		n= gTheSystem.theCEDG.BFSSortedNodes.get_next(n);
	}

#if 0
	pos2 = gTheSystem.compNodeList.GetHeadPosition();
	if(m_cboHeuristic.GetCurSel() == 1) {
		//Use first-fit heuristic
		while(pos) {
			CProcessorNode *pNode = gTheSystem.newProcessorNodeList.GetNext(pos);
			double totalUtil = 0.0;
			//POSITION pos2 = gTheSystem.compNodeList.GetHeadPosition();
			while(pos2) {
				CComponentNode *cNode = gTheSystem.compNodeList.GetNext(pos2);
				//Do not allocate proxy components
				//if(cNode->type == PROXY) continue;
				//If a component is locked, keep it on its original processor
				//dont move it.
				if(cNode->locked && cNode->processor) {
					cNode->processor->AddComponentAllocation(cNode);
					continue;
				}
				if(totalUtil + cNode->util > maxUtil) break;
				pNode->AddComponentAllocation(cNode);
			}   
		}
	} else {
		//Use best-fit heuristic
		int numCompsPerCPU = gTheSystem.compNodeList.GetCount() / m_nCPUs;
		while(pos) {
			CProcessorNode *pNode = gTheSystem.newProcessorNodeList.GetNext(pos);
			double totalUtil = 0.0;
			int count = 0;
			//POSITION pos2 = gTheSystem.compNodeList.GetHeadPosition();
			while(pos2) {
				CComponentNode *cNode = gTheSystem.compNodeList.GetNext(pos2);
				//Do not allocate proxy components
				//if(cNode->type == PROXY) continue;
				//If a component is locked, keep it on its original processor
				//dont move it.
				if(cNode->locked && cNode->processor) {
					cNode->processor->AddComponentAllocation(cNode);
					continue;
				}
				pNode->AddComponentAllocation(cNode);
				if(totalUtil + cNode->util > maxUtil || count++ > numCompsPerCPU) break;
			}
		}
	}
	//Add all remaining components to the last processor..poor guy.
	CProcessorNode *pNode = gTheSystem.newProcessorNodeList.GetTail();
	while(pos2) {
		CComponentNode *cNode = gTheSystem.compNodeList.GetNext(pos2);
		pNode->AddComponentAllocation(cNode);
	}
#endif




//////////////////////////////////////////////////////////////////
	m_strAllocResults.Empty();
	//testing code
	/*
	{
	Node *n = gTheSystem.theCEDG.BFSSortedNodes.get_head();
	m_strAllocResults.Format("%d ", gTheSystem.theCEDG.BFSSortedNodes.get_size());
	while(n) {
		m_strAllocResults += gTheSystem.theCEDG.get_comp_node(n->x)->name;
		m_strAllocResults += "  ";
		n = gTheSystem.theCEDG.BFSSortedNodes.get_next(n);
	}
	UpdateData(FALSE);
	return;
	}
	*/
	pos = gTheSystem.newProcessorNodeList.GetHeadPosition();
	int counter = 0;
	while(pos) {
		CProcessorNode *pNode = gTheSystem.newProcessorNodeList.GetNext(pos);
		POSITION pos2 = pNode->components_allocation.GetHeadPosition();
		m_strAllocResults += pNode->name;
		CString temp;
		temp.Format(":%d components\r\n", pNode->components_allocation.GetCount());
		//temp.Format(":\r\n");
		m_strAllocResults += temp;
		while(pos2) {
			CComponentNode *cNode = pNode->components_allocation.GetNext(pos2);
			//Do not show the timer components. they are artificial, not real components.
			if(cNode->isTimer) continue;
			m_strAllocResults += cNode->name;
			m_strAllocResults += ",";
			if(counter++ >= 4) {
				m_strAllocResults += "\r\n";
				counter = 0;
			}
		}
		m_strAllocResults += "\r\n\r\n";
	}
/*
	pos = gTheSystem.compNodeList.GetHeadPosition();
	while(pos) {
		CComponentNode *cNode = gTheSystem.compNodeList.GetNext(pos);
		if(cNode->processor_allocation == NULL) {
			CString cs;
			cs.Format("Component %s does not have a processor allocation!", cNode->name);
			AfxMessageBox(cs);
		}
	}
*/
//	m_strAllocResults = cedg.getAllocationResultStr();
	if(!bNoFeedback) {
		m_btnUpdateAIF.EnableWindow(true);
	}
	UpdateData(FALSE);
	
	return;
//Forget the grahcut algorithms. dont work.
#if 0
	wdigraph cedg(gTheSystem.theCEDG);
	
	cedg.mincut(numCPUs, maxUtil, m_nNumComponents);

	m_strAllocResults = cedg.getAllocationResultStr();
	
	CWnd *pWnd = AfxGetMainWnd ();

	CFrameWnd* pwndMainFrame = (CFrameWnd*)pWnd;
	CFrameWnd *pwndActiveFrame = pwndMainFrame->GetActiveFrame ();

	CDocument *pDocument = pwndActiveFrame ->GetActiveDocument (); 
	CAvionicsDoc *pDoc = (CAvionicsDoc *) pDocument;

//	cedg.UpdateAIFWithAllocResults(pDoc->backend);
	UpdateData(FALSE);
#endif
}

BOOL CAllocateComponentsDlg::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
	m_nUtil = 100;
	m_nCPUs = gTheSystem.processorNodeList.GetCount();
	m_cboHeuristic.AddString("First Fit");
    m_cboHeuristic.AddString("Best Fit");
//	m_cboHeuristic.AddString("Graph Min Cut");
    m_cboHeuristic.SetCurSel(0);

	UpdateData(FALSE);
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void CAllocateComponentsDlg::OnOK() 
{
	ASSERT(m_nCPUs == gTheSystem.processorNodeList.GetCount());
	CAvionicsDoc *pDoc = (CAvionicsDoc *) GetDocumentPtr();
	BeginWaitCursor();
	UpdateAIFWithAllocResults(pDoc->backend);
	EndWaitCursor();
	m_btnUpdateAIF.EnableWindow(false);
//	CDialog::OnOK();
}
